/*
Copyright (C) 2016, Vershinin Ivan
*/

package ru.vershinin.medical.lelmon;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothServerSocket;
import android.bluetooth.BluetoothSocket;
import android.content.Context;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.util.UUID;

public class BluetoothService {

    // Debugging
    private static final String TAG = "BLUETOOTH_SERVICE";

    // Name for the SDP record when creating server socket
    private static final String NAME_SECURE = "BluetoothChatSecure";
    private static final String NAME_INSECURE = "BluetoothChatInsecure";

    // Unique UUID for this application
    private static final UUID MY_UUID_SECURE =
            UUID.fromString("00001101-0000-1000-8000-00805f9b34fb");
    private static final UUID MY_UUID_INSECURE =
            UUID.fromString("8ce255c0-200a-11e0-ac64-0800200c9a66");

    // Member fields
    private final BluetoothAdapter bsAdapter;
    private final Handler bsHandler;
    private AcceptThread bsSecureAcceptThread;
    private AcceptThread bsInsecureAcceptThread;
    private ConnectThread bsConnectThread;
    private ConnectedThread bsConnectedThread;
    private int bsState;
    private Context bsContext;

    // Constants that indicate the current connection state
    public static final int STATE_NONE = 0;       // we're doing nothing
    public static final int STATE_LISTEN = 1;     // now listening for incoming connections
    public static final int STATE_CONNECTING = 2; // now initiating an outgoing connection
    public static final int STATE_CONNECTED = 3;  // now connected to a remote device

    // Constructor. Prepares a new BluetoothChat session.
    public BluetoothService(Context context, Handler handler) {
        bsContext = context;
        bsAdapter = BluetoothAdapter.getDefaultAdapter();
        bsState = STATE_NONE;
        bsHandler = handler;
    }

    // Set the current state of the chat connection
    private synchronized void setState(int state) {
        Log.d(TAG, "setState() " + bsState + " -> " + state);
        bsState = state;
        // Give the new state to the Handler so the UI Activity can update
        bsHandler.obtainMessage(Constants.MESSAGE_STATE_CHANGE, state, -1).sendToTarget();
    }

    // Return the current connection state.
    public synchronized int getState() {
        return bsState;
    }

    // Start the chat service. Specifically start AcceptThread to begin a
    // session in listening (server) mode. Called by the Activity onResume()
    public synchronized void start() {
        Log.d(TAG, "start");

        // Cancel any thread attempting to make a connection
        if (bsConnectThread != null) {
            bsConnectThread.cancel();
            bsConnectThread = null;
        }

        // Cancel any thread currently running a connection
        if (bsConnectedThread != null) {
            bsConnectedThread.cancel();
            bsConnectedThread = null;
        }

        setState(STATE_LISTEN);

        // Start the thread to listen on a BluetoothServerSocket
        if (bsSecureAcceptThread == null) {
            bsSecureAcceptThread = new AcceptThread(true);
            bsSecureAcceptThread.start();
        }
        if (bsInsecureAcceptThread == null) {
            bsInsecureAcceptThread = new AcceptThread(false);
            bsInsecureAcceptThread.start();
        }
    }

    // Start the ConnectThread to initiate a connection to a remote device.
    public synchronized void connect(BluetoothDevice device, boolean secure) {
        Log.d(TAG, "connect to: " + device);

        // Cancel any thread attempting to make a connection
        if (bsState == STATE_CONNECTING) {
            if (bsConnectThread != null) {
                bsConnectThread.cancel();
                bsConnectThread = null;
            }
        }

        // Cancel any thread currently running a connection
        if (bsConnectedThread != null) {
            bsConnectedThread.cancel();
            bsConnectedThread = null;
        }

        // Start the thread to connect with the given device
        bsConnectThread = new ConnectThread(device, secure);
        bsConnectThread.start();
        setState(STATE_CONNECTING);
    }

    // Start the ConnectedThread to begin managing a Bluetooth connection
    public synchronized void connected(BluetoothSocket socket, BluetoothDevice
            device, final String socketType) {
        Log.d(TAG, "connected, Socket Type:" + socketType);

        // Cancel the thread that completed the connection
        if (bsConnectThread != null) {
            bsConnectThread.cancel();
            bsConnectThread = null;
        }

        // Cancel any thread currently running a connection
        if (bsConnectedThread != null) {
            bsConnectedThread.cancel();
            bsConnectedThread = null;
        }

        // Cancel the accept thread because we only want to connect to one device
        if (bsSecureAcceptThread != null) {
            bsSecureAcceptThread.cancel();
            bsSecureAcceptThread = null;
        }
        if (bsInsecureAcceptThread != null) {
            bsInsecureAcceptThread.cancel();
            bsInsecureAcceptThread = null;
        }

        // Start the thread to manage the connection and perform transmissions
        bsConnectedThread = new ConnectedThread(socket, socketType);
        bsConnectedThread.start();

        // Send the name of the connected device back to the UI Activity
        Message msg = bsHandler.obtainMessage(Constants.MESSAGE_DEVICE_NAME);
        Bundle bundle = new Bundle();
        bundle.putString(Constants.DEVICE_NAME, device.getName());
        msg.setData(bundle);
        bsHandler.sendMessage(msg);

        setState(STATE_CONNECTED);
    }

    // Stop all threads
    public synchronized void stop() {
        Log.d(TAG, "stop");

        if (bsConnectThread != null) {
            bsConnectThread.cancel();
            bsConnectThread = null;
        }

        if (bsConnectedThread != null) {
            bsConnectedThread.cancel();
            bsConnectedThread = null;
        }

        if (bsSecureAcceptThread != null) {
            bsSecureAcceptThread.cancel();
            bsSecureAcceptThread = null;
        }

        if (bsInsecureAcceptThread != null) {
            bsInsecureAcceptThread.cancel();
            bsInsecureAcceptThread = null;
        }
        setState(STATE_NONE);
    }

    // Write to the ConnectedThread in an unsynchronized manner
    public void write(byte[] out) {
        // Create temporary object
        ConnectedThread ct;
        // Synchronize a copy of the ConnectedThread
        synchronized (this) {
            if (bsState != STATE_CONNECTED) return;
            ct = bsConnectedThread;
        }
        // Perform the write unsynchronized
        ct.write(out);
    }

    // Indicate that the connection attempt failed and notify the UI Activity.
    private void connectionFailed() {
        // Send a failure message back to the Activity
        Message msg = bsHandler.obtainMessage(Constants.MESSAGE_TOAST);
        Bundle bundle = new Bundle();
        bundle.putString(Constants.TOAST, bsContext.getResources().getString(R.string.msg_unable_to_connect_device));
        msg.setData(bundle);
        bsHandler.sendMessage(msg);

        // Start the service over to restart listening mode
        BluetoothService.this.start();
    }

    // Indicate that the connection was lost and notify the UI Activity.
    private void connectionLost() {
        // Send a failure message back to the Activity
        Message msg = bsHandler.obtainMessage(Constants.MESSAGE_TOAST);
        Bundle bundle = new Bundle();
        bundle.putString(Constants.TOAST, bsContext.getResources().getString(R.string.msg_connect_device_is_lost));
        msg.setData(bundle);
        bsHandler.sendMessage(msg);

        // Start the service over to restart listening mode
        BluetoothService.this.start();
    }

    // This thread runs while listening for incoming connections. It behaves like a server-side client.
    // It runs until a connection is accepted (or until cancelled).
    private class AcceptThread extends Thread {
        // The local server socket
        private final BluetoothServerSocket atServerSocket;
        private String atSocketType;

        public AcceptThread(boolean secure) {
            BluetoothServerSocket tmp = null;
            atSocketType = secure ? "Secure" : "Insecure";

            // Create a new listening server socket
            try {
                if (secure) {
                    tmp = bsAdapter.listenUsingRfcommWithServiceRecord(NAME_SECURE,
                            MY_UUID_SECURE);
                } else {
                    tmp = bsAdapter.listenUsingInsecureRfcommWithServiceRecord(
                            NAME_INSECURE, MY_UUID_INSECURE);
                }
            } catch (IOException e) {
                Log.e(TAG, "Socket Type: " + atSocketType + " listen() failed", e);
            }
            atServerSocket = tmp;
        }

        public void run() {
            Log.d(TAG, "Socket Type: " + atSocketType + " BEGIN AcceptThread " + this);
            setName("AcceptThread" + atSocketType);
            BluetoothSocket socket = null;
            // Listen to the server socket if we're not connected
            while (bsState != STATE_CONNECTED) {
                if (atServerSocket != null) {
                    try {
                        // This is a blocking call and will only return on a successful connection or an exception
                        socket = atServerSocket.accept();
                    } catch (IOException e) {
                        Log.e(TAG, "Socket Type: " + atSocketType + " accept() failed", e);
                        break;
                    }
                }
                // If a connection was accepted
                if (socket != null) {
                    Log.d(TAG, "............................4");
                    synchronized (BluetoothService.this) {
                        switch (bsState) {
                            case STATE_LISTEN:
                            case STATE_CONNECTING:
                                // Situation normal. Start the connected thread.
                                connected(socket, socket.getRemoteDevice(),
                                        atSocketType);
                                break;
                            case STATE_NONE:
                            case STATE_CONNECTED:
                                // Either not ready or already connected. Terminate new socket.
                                try {
                                    socket.close();
                                } catch (IOException e) {
                                    Log.e(TAG, "Could not close unwanted socket", e);
                                }
                                break;
                        }
                    }
                }
            }
            Log.i(TAG, "END AcceptThread, socket Type: " + atSocketType);

        }

        public void cancel() {
            Log.d(TAG, "Socket Type " + atSocketType + " cancel " + this);
            try {
                if (atServerSocket != null)
                    atServerSocket.close();
            } catch (IOException e) {
                Log.e(TAG, "Socket Type " + atSocketType + " close() of server failed", e);
            }
        }
    }

    // This thread runs while attempting to make an outgoing connection with a device.
    // It runs straight through; the connection either succeeds or fails.
    private class ConnectThread extends Thread {
        private final BluetoothSocket ctSocket;
        private final BluetoothDevice ctDevice;
        private String ctSocketType;

        public ConnectThread(BluetoothDevice device, boolean secure) {
            ctDevice = device;
            BluetoothSocket tmp = null;
            ctSocketType = secure ? "Secure" : "Insecure";

            // Get a BluetoothSocket for a connection with the given BluetoothDevice
            try {
                if (secure) {
                    tmp = device.createRfcommSocketToServiceRecord(
                            MY_UUID_SECURE);
                } else {
                    tmp = device.createInsecureRfcommSocketToServiceRecord(
                            MY_UUID_INSECURE);
                }
            } catch (IOException e) {
                Log.e(TAG, "Socket Type: " + ctSocketType + "create() failed", e);
            }
            ctSocket = tmp;
        }

        public void run() {
            Log.i(TAG, "BEGIN ConnectThread SocketType: " + ctSocketType);
            setName("ConnectThread: " + ctSocketType);

            // Always cancel discovery because it will slow down a connection
            bsAdapter.cancelDiscovery();

            // Make a connection to the BluetoothSocket
            try {
                // This is a blocking call and will only return on a successful connection or an exception
                ctSocket.connect();
            } catch (IOException e) {
                // Close the socket
                try {
                    ctSocket.close();
                } catch (IOException e2) {
                    Log.e(TAG, "unable to close() " + ctSocketType + " socket during connection failure", e2);
                }
                connectionFailed();
                return;
            }

            // Reset the ConnectThread because we're done
            synchronized (BluetoothService.this) {
                bsConnectThread = null;
            }

            // Start the connected thread
            connected(ctSocket, ctDevice, ctSocketType);
        }

        public void cancel() {
            try {
                ctSocket.close();
            } catch (IOException e) {
                Log.e(TAG, "close() of connect " + ctSocketType + " socket failed", e);
            }
        }
    }

    // This thread runs during a connection with a remote device.
    // It handles all incoming and outgoing transmissions.
    private class ConnectedThread extends Thread {
        private final BluetoothSocket contSocket;
        private final InputStream contInStream;
        private final OutputStream contOutStream;

        public ConnectedThread(BluetoothSocket socket, String socketType) {
            Log.d(TAG, "create ConnectedThread: " + socketType);
            contSocket = socket;
            InputStream tmpIn = null;
            OutputStream tmpOut = null;

            // Get the BluetoothSocket input and output streams
            try {
                tmpIn = socket.getInputStream();
                tmpOut = socket.getOutputStream();
            } catch (IOException e) {
                Log.e(TAG, "temp sockets not created", e);
            }

            contInStream = tmpIn;
            contOutStream = tmpOut;
        }

        public void run() {
            Log.i(TAG, "BEGIN ConnectedThread");
            // Keep listening to the InputStream while connected
            while (true) {
                try {
                    int bytes = contInStream.available();
                    if (bytes == 0)
                    {
                        try {
                            Thread.sleep(10);
                        } catch (InterruptedException e) {
                            Log.d(TAG, "Ошибка приостановки потока");
                        }
                        continue;
                    }
                    byte[] buffer = new byte[bytes];  // buffer store for the stream
                    // Read from the InputStream
                    bytes = contInStream.read(buffer);
                    // Send the obtained bytes to the UI Activity
                    bsHandler.obtainMessage(Constants.MESSAGE_READ, bytes, -1, buffer)
                            .sendToTarget();
                } catch (IOException e) {
                    Log.e(TAG, "disconnected", e);
                    connectionLost();
                    break;
                }
            }
        }

        // Write to the connected OutStream.
        public void write(byte[] buffer) {
            try {
                contOutStream.write(buffer);
                // Share the sent message back to the UI Activity
                bsHandler.obtainMessage(Constants.MESSAGE_WRITE, -1, -1, buffer).sendToTarget();
            } catch (IOException e) {
                Log.e(TAG, "Exception during write", e);
            }
        }

        public void cancel() {
            try {
                contSocket.close();
            } catch (IOException e) {
                Log.e(TAG, "close() of connect socket failed", e);
            }
        }
    }
}
